import os

from langchain_openai import ChatOpenAI, AzureChatOpenAI, OpenAIEmbeddings, AzureOpenAIEmbeddings
from langchain.chat_models.base import BaseChatModel
from langchain.embeddings.base import Embeddings


def select_model(provider: str = "openai", model_name: str = "gpt-4o-mini", temperature: float = 0.2) -> BaseChatModel:
    if provider == "azure":
        return AzureChatOpenAI(
            azure_deployment=os.environ.get("AZURE_OPENAI_AGENT_DEPLOYMENT_NAME"),
            openai_api_version=os.environ.get("AZURE_OPENAI_API_VERSION"),
            azure_endpoint=f"https://{os.environ.get('AZURE_OPENAI_API_INSTANCE_NAME')}.openai.azure.com",
            openai_api_key=os.environ.get("AZURE_OPENAI_API_KEY"),
            model_name=model_name,
            temperature=temperature
        )
    else:  # default to OpenAI
        return ChatOpenAI(
            model_name=model_name,
            temperature=temperature,
            openai_api_key=os.environ.get("OPENAI_API_KEY")
        )


def select_embeddings(provider: str = "openai", model_name: str = "text-embedding-3-small") -> Embeddings:
    if provider == "azure":
        return AzureOpenAIEmbeddings(
            azure_deployment=os.environ.get("AZURE_OPENAI_EMBEDDING_DEPLOYMENT_NAME"),
            openai_api_version=os.environ.get("AZURE_OPENAI_API_VERSION"),
            azure_endpoint=f"https://{os.environ.get('AZURE_OPENAI_API_INSTANCE_NAME')}.openai.azure.com",
            openai_api_key=os.environ.get("AZURE_OPENAI_API_KEY"),
            model=model_name
        )
    else:  # default to OpenAI
        return OpenAIEmbeddings(
            model=model_name,
            openai_api_key=os.environ.get("OPENAI_API_KEY")
        )



vds_schema = {
    "FieldBase": {
        "type": "object",
        "description": "Common properties of a Field. A Field represents a column of data in a published datasource",
        "required": [ "fieldCaption" ],
        "properties": {
            "fieldCaption": {
                "type": "string",
                "description": "Either the name of a specific Field in the data source, or, in the case of a calculation, a user-supplied name for the calculation."
            },
            "fieldAlias": {
                "type": "string",
                "description": "An alternative name to give the field. Will only be used in Object format output."
            },
            "maxDecimalPlaces": {
                "type": "integer",
                "description": "The maximum number of decimal places. Any trailing 0s will be dropped. The maxDecimalPlaces value must be greater or equal to 0."
            },
            "sortDirection": {
                "$ref": "#/components/schemas/SortDirection"
            },
            "sortPriority": {
                "type": "integer",
                "description": "To enable sorting on a specific Field provide a sortPriority for that field, and that field will be sorted. The sortPriority provides a ranking of how to sort Fields when multiple Fields are being sorted. The highest priority (lowest number) Field is sorted first. If only 1 Field is being sorted, then any value may be used for sortPriority. SortPriority should be an integer greater than 0."
            }
        }
    },
    "Field": {
        "oneOf": [
            {
                "allOf": [
                    {
                        "$ref": "#/components/schemas/FieldBase"
                    }
                ],
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "fieldCaption": {},
                    "fieldAlias": {},
                    "maxDecimalPlaces": {},
                    "sortDirection": {},
                    "sortPriority": {}
                }
            },
            {
                "allOf": [
                    {
                        "$ref": "#/components/schemas/FieldBase"
                    }
                ],
                "type": "object",
                "required": [
                    "function"
                ],
                "additionalProperties": False,
                "properties": {
                    "function": {
                        "$ref": "#/components/schemas/Function"
                    },
                    "fieldCaption": {},
                    "fieldAlias": {},
                    "maxDecimalPlaces": {},
                    "sortDirection": {},
                    "sortPriority": {}
                }
            },
            {
                "allOf": [
                    {
                        "$ref": "#/components/schemas/FieldBase"
                    }
                ],
                "type": "object",
                "required": [
                    "calculation"
                ],
                "additionalProperties": False,
                "properties": {
                    "calculation": {
                        "type": "string",
                        "description": "A Tableau calculation which will be returned as a Field in the Query"
                    },
                    "fieldCaption": {},
                    "fieldAlias": {},
                    "maxDecimalPlaces": {},
                    "sortDirection": {},
                    "sortPriority": {}
                }
            }
        ]
    },
    "FieldMetadata": {
        "type": "object",
        "description": "Describes a field in the datasource that can be used to create queries.",
        "properties": {
            "fieldName": {
                "type": "string"
            },
            "fieldCaption": {
                "type": "string"
            },
            "dataType": {
                "type": "string",
                "enum": [
                    "INTEGER",
                    "REAL",
                    "STRING",
                    "DATETIME",
                    "BOOLEAN",
                    "DATE",
                    "SPATIAL",
                    "UNKNOWN"
                ]
            },
            "logicalTableId": {
                "type": "string"
            }
        }
    },
    "Filter": {
        "type": "object",
        "description": "A Filter to be used in the Query to filter on the datasource",
        "required": ["field", "filterType"],
        "properties": {
            "field": {
                "$ref": "#/components/schemas/FilterField"
            },
            "filterType": {
                "type": "string",
                "enum": [
                    "QUANTITATIVE_DATE",
                    "QUANTITATIVE_NUMERICAL",
                    "SET",
                    "MATCH",
                    "DATE",
                    "TOP"
                ]
            },
            "context": {
                "type": "boolean",
                "description": "Make the given filter a Context Filter, meaning that it's an independent filter. Any other filters that you set are defined as dependent filters because they process only the data that passes through the context filter",
                "default": False
            }
        },
        "discriminator": {
            "propertyName": "filterType",
            "mapping": {
                "QUANTITATIVE_DATE": "#/components/schemas/QuantitativeDateFilter",
                "QUANTITATIVE_NUMERICAL": "#/components/schemas/QuantitativeNumericalFilter",
                "SET": "#/components/schemas/SetFilter",
                "MATCH": "#/components/schemas/MatchFilter",
                "DATE": "#/components/schemas/RelativeDateFilter",
                "TOP": "#/components/schemas/TopNFilter"
            }
        }
    },
    "FilterField": {
        "oneOf": [
            {
                "required": ["fieldCaption"],
                "additionalProperties": False,
                "properties": {
                    "fieldCaption": {
                        "type": "string",
                        "description": "The caption of the field to filter on"
                    }
                }
            },
            {
                "required": ["fieldCaption", "function"],
                "additionalProperties": False,
                "properties": {
                    "fieldCaption": {
                        "type": "string",
                        "description": "The caption of the field to filter on"
                    },
                    "function": {
                        "$ref": "#/components/schemas/Function"
                    }
                }
            },
            {
                "required": ["calculation"],
                "additionalProperties": False,
                "properties": {
                    "calculation": {
                        "type": "string",
                        "description": "A Tableau calculation which will be used to Filter on"
                    }
                }
            }
        ]
    },
    "Function": {
        "type": "string",
        "description": "The standard set of Tableau aggregations which can be applied to a Field",
        "enum": [
            "SUM",
            "AVG",
            "MEDIAN",
            "COUNT",
            "COUNTD",
            "MIN",
            "MAX",
            "STDEV",
            "VAR",
            "COLLECT",
            "YEAR",
            "QUARTER",
            "MONTH",
            "WEEK",
            "DAY",
            "TRUNC_YEAR",
            "TRUNC_QUARTER",
            "TRUNC_MONTH",
            "TRUNC_WEEK",
            "TRUNC_DAY"
        ]
    },
    "MatchFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/Filter"
            },
            {
                "type": "object",
                "description": "A Filter that can be used to match against String Fields",
                "properties": {
                    "contains": {
                        "type": "string",
                        "description": "Matches when a Field contains this value"
                    },
                    "startsWith": {
                        "type": "string",
                        "description": "Matches when a Field starts with this value"
                    },
                    "endsWith": {
                        "type": "string",
                        "description": "Matches when a Field ends with this value"
                    },
                    "exclude": {
                        "type": "boolean",
                        "description": "When true, the inverse of the matching logic will be used",
                        "default": False
                    }
                }
            }
        ]
    },
    "QuantitativeFilterBase": {
        "allOf": [
            {
                "$ref": "#/components/schemas/Filter"
            },
            {
                "type": "object",
                "required": ["quantitativeFilterType"],
                "properties": {
                    "quantitativeFilterType": {
                        "type": "string",
                        "enum": [ "RANGE", "MIN", "MAX", "ONLY_NULL", "ONLY_NON_NULL" ]
                    },
                    "includeNulls": {
                        "type": "boolean",
                        "description": "Only applies to RANGE, MIN, and MAX Filters. Should nulls be returned or not. If not provided the default is to not include null values"
                    }
                }
            }
        ]
    },
    "QuantitativeNumericalFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/QuantitativeFilterBase"
            }
        ],
        "type": "object",
        "description": "A Filter that can be used to find the minimumn, maximumn or range of numerical values of a Field",
        "properties": {
            "min": {
                "type": "number",
                "description": "A numerical value, either integer or floating point indicating the minimum value to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN"
            },
            "max": {
                "type": "number",
                "description": "A numerical value, either integer or floating point indicating the maximum value to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN"
            }
        }
    },
    "QuantitativeDateFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/QuantitativeFilterBase"
            }
        ],
        "type": "object",
        "description": "A Filter that can be used to find the minimum, maximum or range of date values of a Field",
        "properties": {
            "minDate": {
                "type": "string",
                "format": "date",
                "description": "An RFC 3339 date indicating the earliest date to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN"
            },
            "maxDate": {
                "type": "string",
                "format": "date",
                "description": "An RFC 3339 date indicating the latest date to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN"
            }
        }
    },
    "Query": {
        "description": "The Query is the fundamental interface to Headless BI. It holds the specific semantics to perform against the Data Source. A Query consists of an array of Fields to query against, and an optional array of filters to apply to the query",
        "required": [
            "fields"
        ],
        "type": "object",
        "properties": {
            "fields": {
                "description": "An array of Fields that define the query",
                "type": "array",
                "items": {
                    "$ref": "#/components/schemas/Field"
                }
            },
            "filters": {
                "description": "An optional array of Filters to apply to the query",
                "type": "array",
                "items": {
                    "$ref": "#/components/schemas/Filter"
                }
            }
        },
        "additionalProperties": False
    },
    "SetFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/Filter"
            }
        ],
        "type": "object",
        "description": "A Filter that can be used to filter on a specific set of values of a Field",
        "required": [
            "values"
        ],
        "properties": {
            "values": {
                "type": "array",
                "items": {},
                "description": "An array of values to filter on"
            },
            "exclude": {
                "type": "boolean",
                "default": False
            }
        }
    },
    "SortDirection": {
        "type": "string",
        "description": "The direction of the sort, either ascending or descending. If not supplied the default is ascending",
        "enum": [
            "ASC",
            "DESC"
        ]
    },
    "RelativeDateFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/Filter"
            },
            {
                "type": "object",
                "description": "A Filter that can be used to filter on dates using a specific anchor and fields that specify a relative date range to that anchor",
                "required": ["periodType", "dateRangeType"],
                "properties": {
                    "periodType": {
                        "type": "string",
                        "description": "The units of time in the relative date range",
                        "enum": [
                            "MINUTES",
                            "HOURS",
                            "DAYS",
                            "WEEKS",
                            "MONTHS",
                            "QUARTERS",
                            "YEARS"
                        ]
                    },
                    "dateRangeType": {
                        "type": "string",
                        "description": "The direction in the relative date range",
                        "enum": [
                            "CURRENT",
                            "LAST",
                            "LASTN",
                            "NEXT",
                            "NEXTN",
                            "TODATE"
                        ]
                    },
                    "rangeN": {
                        "type": "integer",
                        "description": "When dateRangeType is LASTN or NEXTN, this is the N value (how many years, months, etc.)."
                    },
                    "anchorDate": {
                        "type": "string",
                        "format": "date",
                        "description": "When this field is not provided, defaults to today."
                    },
                    "includeNulls": {
                        "type": "boolean",
                        "description": "Should nulls be returned or not. If not provided the default is to not include null values"
                    }
                }
            }
        ]
    },
    "TopNFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/Filter"
            },
            {
                "type": "object",
                "description": "A Filter that can be used to find the top or bottom number of Fields relative to the values in the fieldToMeasure",
                "required": ["howMany, fieldToMeasure"],
                "properties": {
                    "direction": {
                        "type": "string",
                        "enum": [
                                "TOP",
                                "BOTTOM"
                            ],
                        "default": "TOP",
                        "description": "Top (Ascending) or Bottom (Descending) N"
                    },
                    "howMany": {
                        "type": "integer",
                        "description": "The number of values from the Top or the Bottom of the given fieldToMeasure"
                    },
                    "fieldToMeasure": {
                        "$ref": "#/components/schemas/FilterField"
                    }
                }
            }
        ]
    }
}

sample_queries = [
    {
        "example": "a simple query",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Category"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ]
        }
    },
     # removing until we can support this with higher reliability
    # {
    #    "example": "a simple Tableau calculation",
    #    "query": {
    #         "fields": [
    #             {
    #                 "fieldCaption": "AOV",
    #                 "calculation": "SUM([Profit])/COUNTD([Order ID])"
    #             }
    #         ]
    #     }
    # },
    {
        "example": "applying a set filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Ship Mode"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Ship Mode"
                    },
                    "filterType": "SET",
                    "values": [ "First Class", "Standard Class" ],
                    "exclude": "false"
                }
            ]
        }
    },
    {
        "example": "applying a quantitative filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Ship Mode"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "RANGE",
                    "min": 266839,
                    "max": 1149562
                }
            ]
        }
    },
    {
        "example": "applying a quantitative date filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Order Date",
                    "function": "YEAR"
                },
                {
                    "fieldCaption": "Order Date",
                    "function": "QUARTER"
                },
                {
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field":{
                        "fieldCaption": "Order Date"
                    },
                    "filterType": "QUANTITATIVE_DATE",
                    "quantitativeFilterType": "MIN",
                    "minDate": "2020-01-01"
                }
            ]
        }
    },
    {
        "example": "applying a date filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Order Date",
                    "function": "YEAR",
                    "sortPriority": 1
                },
                {
                    "fieldCaption": "Order Date",
                    "function": "MONTH",
                    "sortPriority": 2
                },
                {
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "filterType": "DATE",
                    "field": {
                        "fieldCaption": "Order Date"
                    },
                    "periodType": "MONTHS",
                    "dateRangeType": "CURRENT",
                    "anchorDate": "2021-01-01"
                }
            ]
        }
    },
    {
        "example": "applying a match filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "State/Province"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "State/Province"
                    },
                    "filterType": "MATCH",
                    "startsWith": "A",
                    "endsWith": "a",
                    "contains": "o",
                    "exclude": "false"
                }
            ]
        }
    },
    {
        "example": "applying a top N filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "State/Province"
                },{
                    "fieldCaption": "Profit",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "State/Province"
                    },
                    "filterType": "TOP",
                    "howMany": 10,
                    "fieldToMeasure": {
                        "fieldCaption": "Profit",
                        "function": "SUM"
                    },
                    "direction": "TOP"
                }
            ]
        }
    },
    {
        "example": "applying a multi-set filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Ship Mode"
                },
                {
                    "fieldCaption": "Segment"
                }, {
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Ship Mode"
                    },
                    "filterType": "SET",
                    "values": [ "First Class", "Standard Class" ],
                    "exclude": "false"
                }, {
                    "field": {
                        "fieldCaption": "Segment"
                    },
                    "filterType": "SET",
                    "values": [ "Consumer" ],
                    "exclude": "true"
                }
            ]
        }
    },
    {
        "example": "applying multiple quantitative filters",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Ship Mode"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                },{
                    "fieldCaption": "Profit",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "MIN",
                    "min": 266839
                },
                {
                    "field": {
                        "fieldCaption": "Profit",
                        "function": "SUM"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "MAX",
                    "max": 164098
                }
            ]
        }
    },
    {
        "example": "applying set and quantitative filters",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Ship Mode"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Ship Mode"
                    },
                    "filterType": "SET",
                    "values": [ "First Class", "Standard Class" ],
                    "exclude": "true"
                }, {
                    "field": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "MIN",
                    "min": 400000
                }
            ]
        }
    },
    {
        "example": "applying a context filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Sub-Category"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Sub-Category"
                    },
                    "filterType": "TOP",
                    "howMany": 10,
                    "fieldToMeasure": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    },
                    "direction": "TOP"
                },
                {
                    "field": {
                        "fieldCaption": "Category"
                    },
                    "filterType": "SET",
                    "values": [ "Furniture"],
                    "exclude": "false",
                    "context": "true"
                }
            ]
        }
    },
    {
        "example": "applying date, set and quantitative filters",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Order Date",
                    "function": "TRUNC_DAY",
                    "sortPriority": 1,
                    "sortDirection": "ASC"
                },
                {
                    "fieldCaption": "Sales",
                    "function": "SUM"
                },
                {
                    "fieldCaption": "Ship Mode"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "RANGE",
                    "min": 10,
                    "max": 63
                },
                {
                    "filterType": "DATE",
                    "field": {
                        "fieldCaption": "Order Date"
                    },
                    "periodType": "MONTHS",
                    "dateRangeType": "NEXTN",
                    "rangeN": 3,
                    "anchorDate": "2021-01-01"
                },
                {
                    "field": {
                        "fieldCaption": "Ship Mode"
                    },
                    "filterType": "SET",
                    "values": [ "First Class"],
                    "exclude": "false"
                }
            ]
        }
    },
    {
        "example": "Filtering data to a specific date using DATE filter, dates shown with TRUNC_DAY for day level accuracy, anchorDate is optional and if left empty defaults to today",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Sales",
                    "function": "SUM"
                },
                {
                    "fieldCaption": "Order Date",
                    "function": "TRUNC_DAY",
                    "sortPriority": 1,
                    "sortDirection": "ASC"
                }
            ],
            "filters": [
                {
                    "filterType": "DATE",
                    "field": {
                        "fieldCaption": "Date"
                    },
                    "periodType": "DAYS",
                    "dateRangeType": "CURRENT",
                    "anchorDate": "2021-01-01"
                }
            ]
        }
    },
    {
        "example": "Relative DATE filter to handle questions about last 2 weeks where rangeN is used, anchorDate is optional and if left empty defaults to today",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Sales",
                    "function": "SUM"
                },
                {
                    "fieldCaption": "Orders",
                    "function": "SUM"
                },
                {
                    "fieldCaption": "Order Date",
                    "function": "TRUNC_DAY",
                    "sortPriority": 1,
                    "sortDirection": "ASC"
                },
            ],
            "filters": [
                {
                    "filterType": "DATE",
                    "field": {
                        "fieldCaption": "Order Date"
                    },
                    "periodType": "WEEKS",
                    "dateRangeType": "LASTN",
                    "rangeN": 2,
                    "anchorDate": "2025-02-22"
                }
            ]
        }
    },
    {
        "example": "Relative DATE filter to handle questions about last week, rangeN is not used here, anchorDate is optional and if left empty defaults to today",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Sales",
                    "function": "SUM"
                },
                {
                    "fieldCaption": "Orders",
                    "function": "SUM"
                },
                {
                    "fieldCaption": "Order Date",
                    "function": "TRUNC_DAY",
                    "sortPriority": 1,
                    "sortDirection": "ASC"
                },
            ],
            "filters": [
                {
                    "filterType": "DATE",
                    "field": {
                        "fieldCaption": "Order Date"
                    },
                    "periodType": "WEEKS",
                    "dateRangeType": "LAST"
                }
            ]
        }
    }
]

error_queries = [
    {
        "observation": "ERROR when querying data for a specific date, you need to use RANGE with minDate and maxDate on the same field",
        "error": "Cannot have multiple Filters for the same Field, or the same Field with the same function",
        "error_query": {
            "fields": [
                {
                "fieldCaption": "Orders",
                "function": "SUM"
                }
            ],
            "filters": [
                {
                "field": {
                    "fieldCaption": "Date"
                },
                "filterType": "QUANTITATIVE_DATE",
                "quantitativeFilterType": "MIN",
                "minDate": "2025-01-20"
                },
                {
                "field": {
                    "fieldCaption": "Date"
                },
                "filterType": "QUANTITATIVE_DATE",
                "quantitativeFilterType": "MAX",
                "maxDate": "2025-01-20"
                }
            ]
        },
        "correction": {
            "fields": [
                {
                "fieldCaption": "Orders",
                "function": "SUM"
                }
            ],
            "filters": [
                {
                "field": {
                    "fieldCaption": "Date"
                },
                "filterType": "QUANTITATIVE_DATE",
                "quantitativeFilterType": "RANGE",
                "minDate": "2025-01-20",
                "maxDate": "2025-01-20"
                }
            ]
        }
    },
    {
        "observation": "ERROR when applying `sortDirection` to the entire payload, these properties only apply to fields",
        "error": "Error at 'query': Additional property 'sortDirection' is not allowed",
        "error_query": {
            "fields": [
                {
                    "fieldCaption":"Order Date",
                    "function":"YEAR"
                },
                {
                    "fieldCaption":"Orders",
                    "function":"SUM"
                }
            ],
            "filters":[],
            "sortDirection":"ASC",
            "sortPriority":1
        },
        "correction": {
            "fields": [
                {
                    "fieldCaption":"Order Date",
                    "function":"YEAR",
                    "sortDirection":"ASC",
                    "sortPriority":1
                },
                {
                    "fieldCaption":"Orders",
                    "function":"SUM"
                }
            ],
            "filters":[]
        }
    },
    {
        "observation": "ERROR caused by non-existant property",
        "error": "Error at 'query.filters.0.filterType': Value 'RELATIVE_DATE' is not defined in the schema",
        "error_query": {
            "fields": [
                {
                "fieldCaption": "Order Date",
                "function": "TRUNC_DAY",
                "sortDirection": "ASC",
                "sortPriority": 1
                },
                {
                "fieldCaption": "Profit",
                "function": "SUM"
                },
            ],
            "filters": [
                {
                "field": {
                    "fieldCaption": "Order Date"
                },
                "filterType": "RELATIVE_DATE",
                "periodType": "WEEKS",
                "dateRangeType": "LASTN",
                "rangeN": 5
                }
            ]
        },
        "correction": {
            "fields": [
                {
                "fieldCaption": "Order Date",
                "function": "TRUNC_DAY",
                "sortDirection": "ASC",
                "sortPriority": 1
                },
                {
                "fieldCaption": "Profit",
                "function": "SUM"
                },
            ],
            "filters": [
                {
                "field": {
                    "fieldCaption": "Order Date"
                },
                "filterType": "DATE",
                "periodType": "WEEKS",
                "dateRangeType": "LAST",
                }
            ]
        }
    }
]

vds_prompt_data = {
    "task": {},
    "meta": {},
    "data_dictionary": {},
    "data_model": {},
    "vds_schema": vds_schema,
    "sample_queries": sample_queries,
    "error_queries": error_queries,
    "previous_call_error": {},
    "previous_vds_payload": {}
}

vds_query = """
Task:
Your job is to write the main body of a request to the Tableau VizQL Data Service (VDS) API to
obtain data that answers the task given to you by the user:

User Task: {task}

Data Dictionary:
Use this to map the user's natural language questions to the fields of data available in the data source and
to be aware of any additional operations that may be needed to conceptualize the data correctly according to business
semantics or other logic such as applying filters, aggregations, dates, etc.

{data_dictionary}

Data Model:
Provides sample values for fields in the data source. This is useful in particular when aggregating or inferring
filter values

{data_model}

VDS Schema:
OpenAPI schema describing JSON payloads to the VDS API, use this to generate queries with correct syntax

{vds_schema}

Query:
The query must be written according to the `vds_schema.Query` key. Which describes two properties: fields (required)
and filters (optional)

Fields:
To satisfy the required "fields" property of `vds_schema.Query`, add fields according to the `vds_schema.Field` key,
which references `vds_schema.FieldBase`. Use the `data_dictionary` and `data_model` keys to query all useful or related
fields, including those not directly related to the topics mentioned by the user. Even if additional transformations or
calculations are needed, the additional fields may be useful. DO NOT HALLUCINATE FIELD NAMES

Aggregations:
Aggregations are a property of `vds_schema.Field` called "functions" and are described in `vds_schema.Functions`.
For INTEGER or REAL fields, you must always aggregate it with one of these: SUM, AVG, MEDIAN, COUNT, COUNTD, MIN or MAX.
For DATETIME or DATE fields, you must always aggregate it with one of these: YEAR, QUARTER, MONTH, WEEK, DAY, TRUNC_YEAR,
TRUNC_QUARTER, TRUNC_MONTH, TRUNC_WEEK or TRUNC_DAY. If you get an error from VDS that the response size is too large,
try further aggregating or filtering the data to avoid row-level results that are too granular and not insightful.

Sorting:
Sort fields as often as possible to highlight data of interest in the query even if not explicitly stated by the user. That
means that if they asked about a field in particular, find a way to sort it that makes sense. Sorting is composed of two
properties applied to `vds_schema.Field`: "sortDirection" described by `vds_schema.SortDirection` and "SortPriority" which
is sets the sort order for fields in the query. "SortPriority" is only needed for fields you wish to sort. DO NOT apply
sorting to the entire query or payload, this applies only to fields

Filtering:
Add filters to narrow down the data set according to user specifications and to avoid unnecessary large volumes of data.
Filters are the second and optional property of `vds_schema.Query` and should be written according to `vds_schema.Filter`.
The `vds_schema.Filter` spec references `vds_schema.FilterField`. When asked about values for a specific date, use
QuantitativeDateFilter with RANGE and always include both minDate and maxDate properties. When asked about last week,
previous month, current year, this quarter, previous 10 years, last 2 quarters use `RelativeDateFilter`

There are many types of filters. To choose the right kind of filters you must first use the `data_model` key to map the
target field to the kind of filters it supports. Use the "dataType" for each field (ex. "dataType": "STRING") and the
following list of filter types to make this determination:

- MatchFilter (defined at `vds_schema.MatchFilter`):
- QuantitativeFilterBase (defined at `vds_schema.QuantitativeFilterBase`):
- QuantitativeNumericalFilter (defined at `vds_schema.QuantitativeNumericalFilter`):
- QuantitativeDateFilter (defined at `vds_schema.QuantitativeDateFilter`): Always include minDate and maxDate properties for
specific dates
- SetFilter (defined at `vds_schema.SetFilter`):
- RelativeDateFilter (defined at `vds_schema.RelativeDateFilter`): Ideal for relative dates such as last week, previous month,
current year, this quarter, previous 10 years, last 2 quarters
- TopNFilter (defined at `vds_schema.TopNFilter`): Use this filter when the user asked a Top 10 or Top N question so that
you filter the data response to analyze

You may not have all filter members for fields of type "STRING" in the Data Model, only sample values. Therefore, you must
generate educated guesses for actual filter values and use any previous empty array errors to retry with better values

Sample Queries:
Reference these examples as best practices to execute tasks. These examples show distinct ways to interact with the VDS API
in order to obtain data in different shapes.

{sample_queries}

Error Queries:
These examples demonstrate common errors you have generated in the past, avoid these scenarios by using correct syntax instead

{error_queries}

Previous Tool Call Errors:
If this section has data, then the previous attempt resulted in an error described here:

{previous_call_error}

If the array was empty without syntax errors this indicates that a filter was applied with an incorrect value

The query you generated that caused the error is this:

{previous_vds_payload}

Output:
Your output must be minimal, containing only the VDS query in JSON format without any extra formatting for readability.
If the data source does not contain fields of data that can answer the user_input, return a message so the agent knows to
use a different tool
"""

vds_response = """
This is the output of a data query tool used to fetch information via Tableau's VizQL API
Your task is to synthesize all of this information to provide a clear, concise answer to the end user.

Data Source Name: {data_source_name}
Description: {data_source_description}
Maintainer: {data_source_maintainer}

This is the query written to Tableau's VizQL API:
{vds_query}

This is the resulting data from the query:
{data_table}

This was the user_input (question or task):
{user_input}

Based on the provided context, formulate a comprehensive and informative response to the user's query.
Your response should:
1. Describe the data source name, description and maintainer if this is the first interaction the user has with it
2. Use the resulting data to answer the user's question or task
3. Be short and concise, if the data table is too long only return the relevant rows or a small sample

Your synthesized response:
"""



from typing import Optional
from pydantic import BaseModel, Field

from langchain.prompts import PromptTemplate
from langchain_core.tools import tool, ToolException

from langchain_tableau.tools.prompts import vds_query, vds_prompt_data, vds_response
from langchain_tableau.utilities.auth import jwt_connected_app
from langchain_tableau.utilities.models import select_model
from langchain_tableau.utilities.simple_datasource_qa import (
    env_vars_simple_datasource_qa,
    augment_datasource_metadata,
    get_headlessbi_data,
    prepare_prompt_inputs
)


class DataSourceQAInputs(BaseModel):
    """Describes inputs for usage of the simple_datasource_qa tool"""

    user_input: str = Field(
        ...,
        description="""Describe the user query thoroughly in natural language such as: 'orders and sales for April 20 2025'.
        You can ask for relative dates such as last week, 3 days ago, current year, previous 3 quarters or
        specific dates: profits and average discounts for last week""",
        examples=[
            "sales and orders for April 20 2025"
        ]
    )
    previous_call_error: Optional[str] = Field(
        None,
        description="""If the previous interaction resulted in a VizQL Data Service error, include the error otherwise use None:
        Error: Quantitative Filters must have a QuantitativeFilterType""",
        examples=[
            None, # no errors example
            "Error: Quantitative Filters must have a QuantitativeFilterType"
        ],
    )
    previous_vds_payload: Optional[str] = Field(
        None,
        description="""If the previous interaction resulted in a VizQL Data Service error, include the faulty VDS JSON payload
        otherwise use None: {\"fields\":[{\"fieldCaption\":\"Sub-Category\",\"fieldAlias\":\"SubCategory\",\"sortDirection\":\"DESC\",
        \"sortPriority\":1},{\"function\":\"SUM\",\"fieldCaption\":\"Sales\",\"fieldAlias\":\"TotalSales\"}],
        \"filters\":[{\"field\":{\"fieldCaption\":\"Order Date\"},\"filterType\":\"QUANTITATIVE_DATE\",\"minDate\":\"2023-04-01\",
        \"maxDate\":\"2023-10-01\"},{\"field\":{\"fieldCaption\":\"Sales\"},\"filterType\":\"QUANTITATIVE_NUMERICAL\",
        \"quantitativeFilterType\":\"MIN\",\"min\":200000},{\"field\":{\"fieldCaption\":\"Sub-Category\"},\"filterType\":\"MATCH\",
        \"exclude\":true,\"contains\":\"Technology\"}]}""",
        examples=[
            None, # no errors example
            "{\"fields\":[{\"fieldCaption\":\"Sub-Category\",\"fieldAlias\":\"SubCategory\",\"sortDirection\":\"DESC\",\"sortPriority\":1},{\"function\":\"SUM\",\"fieldCaption\":\"Sales\",\"fieldAlias\":\"TotalSales\"}],\"filters\":[{\"field\":{\"fieldCaption\":\"Order Date\"},\"filterType\":\"QUANTITATIVE_DATE\",\"minDate\":\"2023-04-01\",\"maxDate\":\"2023-10-01\"},{\"field\":{\"fieldCaption\":\"Sales\"},\"filterType\":\"QUANTITATIVE_NUMERICAL\",\"quantitativeFilterType\":\"MIN\",\"min\":200000},{\"field\":{\"fieldCaption\":\"Sub-Category\"},\"filterType\":\"MATCH\",\"exclude\":true,\"contains\":\"Technology\"}]}"
        ],
    )


def initialize_simple_datasource_qa(
    domain: Optional[str] = None,
    site: Optional[str] = None,
    jwt_client_id: Optional[str] = None,
    jwt_secret_id: Optional[str] = None,
    jwt_secret: Optional[str] = None,
    tableau_api_version: Optional[str] = None,
    tableau_user: Optional[str] = None,
    datasource_luid: Optional[str] = None,
    model_provider: Optional[str] = None,
    tooling_llm_model: Optional[str] = None
):
    """
    Initializes the Langgraph tool called 'simple_datasource_qa' for analytical
    questions and answers on a Tableau Data Source

    Args:
        domain (Optional[str]): The domain of the Tableau server.
        site (Optional[str]): The site name on the Tableau server.
        jwt_client_id (Optional[str]): The client ID for JWT authentication.
        jwt_secret_id (Optional[str]): The secret ID for JWT authentication.
        jwt_secret (Optional[str]): The secret for JWT authentication.
        tableau_api_version (Optional[str]): The version of the Tableau API to use.
        tableau_user (Optional[str]): The Tableau user to authenticate as.
        datasource_luid (Optional[str]): The LUID of the data source to perform QA on.
        tooling_llm_model (Optional[str]): The LLM model to use for tooling operations.

    Returns:
        function: A decorated function that can be used as a langgraph tool for data source QA.

    The returned function (datasource_qa) takes the following parameters:
        user_input (str): The user's query or command represented in simple SQL.
        previous_call_error (Optional[str]): Any error from a previous call, for error handling.

    It returns a dictionary containing the results of the QA operation.

    Note:
        If arguments are not provided, the function will attempt to read them from
        environment variables, typically stored in a .env file.
    """
    # if arguments are not provided, the tool obtains environment variables directly from .env
    env_vars = env_vars_simple_datasource_qa(
        domain=domain,
        site=site,
        jwt_client_id=jwt_client_id,
        jwt_secret_id=jwt_secret_id,
        jwt_secret=jwt_secret,
        tableau_api_version=tableau_api_version,
        tableau_user=tableau_user,
        datasource_luid=datasource_luid,
        model_provider=model_provider,
        tooling_llm_model=tooling_llm_model
    )

    @tool("simple_datasource_qa", args_schema=DataSourceQAInputs)
    def simple_datasource_qa(
        user_input: str,
        previous_call_error: Optional[str] = None,
        previous_vds_payload: Optional[str] = None
    ) -> dict:
        """
        Queries a Tableau data source for analytical Q&A. Returns a data set you can use to answer user questions.
        To be more efficient, describe your entire query in a single request rather than selecting small slices of
        data in multiple requests. DO NOT perform multiple queries if all the data can be fetched at once with the
        same filters or conditions:

        Good query: "Profits & average discounts by region for last week"
        Bad queries: "profits per region last week" & "average discounts per region last week"

        If you received an error after using this tool, mention it in your next attempt to help the tool correct itself.
        """

        # Session scopes are limited to only required authorizations to Tableau resources that support tool operations
        access_scopes = [
            "tableau:content:read", # for quering Tableau Metadata API
            "tableau:viz_data_service:read" # for querying VizQL Data Service
        ]
        try:
            tableau_session = jwt_connected_app(
                tableau_domain=env_vars["domain"],
                tableau_site=env_vars["site"],
                jwt_client_id=env_vars["jwt_client_id"],
                jwt_secret_id=env_vars["jwt_secret_id"],
                jwt_secret=env_vars["jwt_secret"],
                tableau_api=env_vars["tableau_api_version"],
                tableau_user=env_vars["tableau_user"],
                scopes=access_scopes
            )
        except Exception as e:
            auth_error_string = f"""
            CRITICAL ERROR: Could not authenticate to the Tableau site successfully.
            This tool is unusable as a result.
            Error from remote server: {e}

            INSTRUCTION: Do not ask the user to provide credentials directly or in chat since they should
            originate from a secure Connected App or similar authentication mechanism. You may inform the
            user that you are not able to access their Tableau environment at this time. You can also describe
            the nature of the error to help them understand why you can't service their request.
            """
            raise ToolException(auth_error_string)

        # credentials to access Tableau environment on behalf of the user
        tableau_auth =  tableau_session['credentials']['token']

        # Data source for VDS querying
        tableau_datasource = env_vars["datasource_luid"]

        # 0. Obtain metadata about the data source to enhance the query writing prompt
        query_writing_data = augment_datasource_metadata(
            task = user_input,
            api_key = tableau_auth,
            url = domain,
            datasource_luid = tableau_datasource,
            prompt = vds_prompt_data,
            previous_errors = previous_call_error,
            previous_vds_payload = previous_vds_payload
        )

        # 1. Insert instruction data into the template
        query_writing_prompt = PromptTemplate(
            input_variables=[
                "task"
                "instructions",
                "vds_schema",
                "sample_queries",
                "error_queries",
                "data_dictionary",
                "data_model",
                "previous_call_error",
                "previous_vds_payload"
            ],
            template=vds_query
        )

        # 2. Instantiate language model to execute the prompt to write a VizQL Data Service query
        query_writer = select_model(
            provider=env_vars["model_provider"],
            model_name=env_vars["tooling_llm_model"],
            temperature=0
        )

        # 3. Query data from Tableau's VizQL Data Service using the AI written payload
        def get_data(vds_query):
            payload = vds_query.content
            try:
                data = get_headlessbi_data(
                    api_key = tableau_auth,
                    url = domain,
                    datasource_luid = tableau_datasource,
                    payload = payload
                )

                return {
                    "vds_query": payload,
                    "data_table": data,
                }
            except Exception as e:
                query_error_message = f"""
                Tableau's VizQL Data Service return an error for the generated query:

                {str(vds_query.content)}

                The user_input used to write this query was:

                {str(user_input)}

                This was the error:

                {str(e)}

                Consider retrying this tool with the same inputs but include the previous query
                causing the error and the error itself for the tool to correct itself on a retry.
                If the error was an empty array, this usually indicates an incorrect filter value
                was applied, thus returning no data
                """

                raise ToolException(query_error_message)

        # 4. Prepare inputs for a structured response to the calling Agent
        def response_inputs(input):
            metadata = query_writing_data.get('meta')
            data = {
                "query": input.get('vds_query', ''),
                "data_source_name": metadata.get('datasource_name'),
                "data_source_description": metadata.get('datasource_description'),
                "data_source_maintainer": metadata.get('datasource_owner'),
                "data_table": input.get('data_table', ''),
            }
            inputs = prepare_prompt_inputs(data=data, user_string=user_input)
            return inputs

        # 5. Response template for the Agent with further instructions
        response_prompt = PromptTemplate(
            input_variables=[
                "data_source_name",
                "data_source_description",
                "data_source_maintainer",
                "vds_query",
                "data_table",
                "user_input"
            ],
            template=vds_response
        )

        # this chain defines the flow of data through the system
        chain = query_writing_prompt | query_writer | get_data | response_inputs | response_prompt


        # invoke the chain to generate a query and obtain data
        vizql_data = chain.invoke(query_writing_data)

        # Return the structured output
        return vizql_data

    return simple_datasource_qa
